from typing import cast

from fastapi import Request

from ..logging import get_logger
from ..settings import Settings
from .models import User

settings = Settings()


class LoginForm:
    """
    The form class for the login form.
    """

    def __init__(self, request: Request) -> None:
        #: The current request object
        self.request: Request = request
        #: The errors to display to the user
        self.errors: list = []
        #: The username to authenticate
        self.username: str | None = None
        #: The password to authenticate
        self.password: str | None = None
        #: The service to redirect to after authentication
        self.service: str = "/"
        #: The title of the site
        self.site_title: str = settings.auth_realm

    async def load_data(self) -> None:
        """
        Load data from request form.
        """
        form = await self.request.form()
        self.username = cast("str", form.get("username"))
        self.password = cast("str", form.get("password"))
        self.service = cast("str", form.get("service", "/"))

    async def is_valid(self) -> bool:
        """
        Return whether the form is valid.

        * the form must have a non-empty username and password
        * the user must exist in LDAP meaning that the user must be in the
          results of the ldap search
          named by :py:attr:`nginx_ldap_auth.settings.Settings.ldap_get_user_filter`
        * If the ``X-Authorization-Filter`` header or
          :py:attr:`nginx_ldap_auth.settings.Settings.ldap_authorization_filter`
          is not ``None``, the user must be in the results of that LDAP search
          The optional header will override the setting.
        * the bind to LDAP must be successful

        If all those tests pass, return ``True``.  Otherwise, return ``False``.

        Returns:
            ``True`` if the form is valid, ``False`` otherwise.

        """
        _logger = get_logger(self.request)
        if not self.username:
            _logger.info("auth.failed.no_username")
            self.errors.append("Username is required")
        if not self.password:
            _logger.info("auth.failed.no_password")
            self.errors.append("A valid password is required")
        if user := await User.objects.get(cast("str", self.username)):
            # The user exists in LDAP
            user = cast("User", user)
            # Ensure that the user is authorized to access this service
            ldap_authorization_filter: str = self.request.headers.get("X-Authorization-Filter", settings.ldap_authorization_filter)
            if not await User.objects.is_authorized(cast("str", self.username), ldap_authorization_filter):
                self.errors.append("You are not authorized to access this service.")
                _logger.warning(
                    "auth.failed.not_authorized",
                    username=self.username,
                    full_name=user.full_name,
                    ldap_url=settings.ldap_uri,
                    target=self.service,
                )
            # Now try to authenticate the user
            if await user.authenticate(cast("str", self.password)):
                # The user has provided valid credentials
                _logger.info(
                    "auth.success",
                    username=self.username,
                    full_name=user.full_name,
                    ldap_url=settings.ldap_uri,
                    target=self.service,
                )
            else:
                self.errors.append("Invalid username or password.")
                _logger.info(
                    "auth.failed.invalid_credentials",
                    username=self.username,
                    target=self.service,
                )
        else:
            self.errors.append("Invalid username or password.")
            _logger.warning(
                "auth.failed.no_such_user",
                username=self.username,
                target=self.service,
            )
        return bool(not self.errors)



from typing import ClassVar, Optional, cast

import bonsai
from bonsai.errors import (
    AuthenticationError,
    LDAPError,
)
from bonsai.utils import escape_filter_exp
from pydantic import BaseModel

from nginx_ldap_auth.ldap import (
    TimeLimitedAIOConnectionPool,
    TimeLimitedAIOLDAPConnection,
)
from nginx_ldap_auth.logging import logger
from nginx_ldap_auth.settings import Settings
from nginx_ldap_auth.types import LDAPObject


class UserManager:
    """
    Manage users in the LDAP directory.
    """

    #: The model class for users
    model: ClassVar[type["User"]]

    def __init__(self) -> None:
        #: The application settings
        self.settings = Settings()
        #: The LDAP connection pool
        self.pool: TimeLimitedAIOConnectionPool | None = None

    def client(self) -> bonsai.LDAPClient:
        """
        Return a new LDAP client instance.

        If :py:attr:`nginx_ldap_auth.settings.Settings.ldap_starttls` is ``True``,
        the client will be configured to use TLS.
        """
        client = bonsai.LDAPClient(
            cast("str", self.settings.ldap_uri), tls=self.settings.ldap_starttls
        )
        client.set_cert_policy("never")
        client.set_ca_cert(None)
        client.set_ca_cert_dir(None)
        client.ignore_referrals = self.settings.ldap_disable_referrals
        client.set_server_chase_referrals(not self.settings.ldap_disable_referrals)
        client.set_async_connection_class(TimeLimitedAIOLDAPConnection)
        return client

    async def create_pool(self) -> None:
        """
        Create the LDAP connection pool and save it as :py:attr:`pool`.
        """
        client = self.client()
        if self.settings.ldap_binddn and self.settings.ldap_password:
            client.set_credentials(
                "SIMPLE",
                user=self.settings.ldap_binddn,
                password=self.settings.ldap_password,
            )
        self.pool = TimeLimitedAIOConnectionPool(
            self.settings,
            client,
            minconn=self.settings.ldap_min_pool_size,
            maxconn=self.settings.ldap_max_pool_size,
            expires=self.settings.ldap_pool_connection_lifetime_seconds,
        )
        await self.pool.open()

    async def authenticate(self, username: str, password: str) -> bool:
        """
        Authenticate a user against the LDAP server.

        If :py:attr:`nginx_ldap_auth.settings.Settings.ldap_user_basedn` is set,
        we will prepend the username with that value to create the DN to bind
        with like so: "{username}{ldap_user_base_dn}.  Otherwise, we will use
        the value of
        :py:attr:`nginx_ldap_auth.settings.Settings.ldap_username_attribute` to
        create the DN as ``{username_attribute}={username},{ldap_basedn}``.

        Args:
            username: the username to authenticate
            password: the password to authenticate with

        Raises:
            LDAPError: if an error occurs while communicating with the LDAP server

        Returns:
            ``True`` if the user is authenticated, ``False`` otherwise

        """
        if self.settings.ldap_user_basedn:
            # This is AD and we need to use the userPrincipalName
            dn = f"{username}{self.settings.ldap_user_basedn}"
        else:
            dn = (
                f"{self.settings.ldap_username_attribute}={username},"
                f"{self.settings.ldap_basedn}"
            )
        client = self.client()
        client.set_credentials("SIMPLE", user=dn, password=password)
        logger.info(
            "ldap.authenticate",
            dn=dn,
            uri=self.settings.ldap_uri,
        )
        try:
            await client.connect(is_async=True)
        except AuthenticationError as e:
            logger.error(
                "ldap.authenticate.error.invalid_credentials",
                dn=dn,
                uri=self.settings.ldap_uri,
                exc_info=str(e),
            )
            return False
        except LDAPError:
            logger.exception("ldap.authenticate.exception", uid=username)
            raise
        return True

    async def exists(self, username: str) -> bool:
        """
        Return ``True`` if the user exists in the LDAP directory, ``False``
        otherwise.

        Args:
            username: the username to check

        Raises:
            LDAPError: if an error occurred while communicating with the LDAP server
            AuthenticationError: if the LDAP server rejects the credentials of
                :py:class:`nginx_ldap_auth.settings.Settings.ldap_binddn` and
                :py:class:`nginx_ldap_auth.settings.Settings.ldap_password`

        Returns:
            ``True`` if the user exists in the LDAP directory, ``False``
            otherwise

        """
        return await self.get(username) is not None

    async def is_authorized(self, username: str, ldap_authorization_filter: str | None) -> bool:
        """
        Test whether the user is authorized to log in.  This is done by
        performing an LDAP search using the filter specified in a header or
        :py:class:`nginx_ldap_auth.settings.Settings.ldap_authorization_filter`.
        If the value is ``None``, the user is considered authorized.

        Args:
            username: the username to check
            ldap_authorization_filter: LDAP authorization filter (optional)

        Raises:
            LDAPError: if an error occurred while communicating with the LDAP server
            AuthenticationError: if the LDAP server rejects the credentials of
                :py:class:`nginx_ldap_auth.settings.Settings.ldap_binddn` and
                :py:class:`nginx_ldap_auth.settings.Settings.ldap_password`

        Returns:
            ``True`` if the user is authorized to log in, ``False`` otherwise.

        """
        if not self.pool:
            await self.create_pool()
        pool = cast("TimeLimitedAIOConnectionPool", self.pool)
        logger.debug(
            "ldap.is_authorized",
            username=username,
            ldap_authorization_filter=ldap_authorization_filter,
        )
        if ldap_authorization_filter is None:
            return True
        try:
            async with pool.spawn() as conn:
                results = await conn.search(
                    base=self.settings.ldap_basedn,
                    scope=bonsai.LDAPSearchScope.SUBTREE,
                    filter_exp=ldap_authorization_filter.format(
                        username_attribute=self.settings.ldap_username_attribute,
                        fullname_attribute=self.settings.ldap_full_name_attribute,
                        username=escape_filter_exp(username),
                    ),
                    attrlist=[self.settings.ldap_username_attribute],
                )
        except AuthenticationError:
            logger.error(
                "ldap.is_authorized.error.invalid_credentials",
                bind_dn=self.settings.ldap_binddn,
            )
            raise
        except LDAPError:
            logger.exception(
                "ldap.is_authorized.exception",
                bind_dn=self.settings.ldap_binddn,
                username=username,
            )
            raise
        return len(results) > 0

    async def get(self, username: str) -> Optional["User"]:
        """
        Get a user from the LDAP directory, and return it as a :py:class:`User`.
        When getting the user, we will use the LDAP search filter specified in
        :py:class:`nginx_ldap_auth.settings.Settings.ldap_get_user_filter`.

        Args:
            username: the username for which to get user information

        Raises:
            LDAPError: if an error occurred while communicating with the LDAP server
            AuthenticationError: if the LDAP server rejects the credentials of
                :py:class:`nginx_ldap_auth.settings.Settings.ldap_binddn` and
                :py:class:`nginx_ldap_auth.settings.Settings.ldap_password`

        Returns:
            The user information as a :py:class:`User` instance, or ``None`` if
            the user is not returned by the LDAP search filter

        """
        if not self.pool:
            await self.create_pool()
        pool = cast("TimeLimitedAIOConnectionPool", self.pool)
        try:
            async with pool.spawn() as conn:
                results = await conn.search(
                    base=self.settings.ldap_basedn,
                    scope=bonsai.LDAPSearchScope.SUBTREE,
                    filter_exp=self.settings.ldap_get_user_filter.format(
                        username_attribute=self.settings.ldap_username_attribute,
                        fullname_attribute=self.settings.ldap_full_name_attribute,
                        username=escape_filter_exp(username),
                    ),
                    attrlist=[
                        self.settings.ldap_username_attribute,
                        self.settings.ldap_full_name_attribute,
                    ],
                )
        except AuthenticationError:
            logger.error(
                "ldap.get_user.error.invalid_credentials",
                bind_dn=self.settings.ldap_binddn,
            )
            raise
        except LDAPError:
            logger.exception(
                "ldap.get_user.exception",
                bind_dn=self.settings.ldap_binddn,
                username=username,
            )
            raise
        if results:
            if len(results) > 1:
                logger.warning(
                    "ldap.get_user.error.multiple_results",
                    bind_dn=self.settings.ldap_binddn,
                    username=username,
                    dns=";".join([r[0] for r in results]),
                )
            return self.model.parse_ldap(results[0])
        return None

    async def cleanup(self) -> None:
        """
        Close the LDAP connection pool.
        """
        if self.pool:
            await self.pool.close()


class User(BaseModel):
    """
    Used to represent a user in the LDAP directory.  It is constructed from the
    LDAP response, and is used to authenticate the user against the LDAP server.
    """

    objects: ClassVar["UserManager"] = UserManager()

    #: The username of the user.
    uid: str
    #: The full name of the user.  We really only use this for logging.
    full_name: str

    async def authenticate(self, password: str) -> bool:
        """
        Authenticate this user against the LDAP server.

        Args:
            password: the password to authenticate with

        Returns:
            ``True`` if the user is authenticated, ``False`` otherwise

        """
        return await self.objects.authenticate(self.uid, password)

    @classmethod
    def parse_ldap(cls, data: LDAPObject) -> "User":
        """
        Parse the LDAP response, and extract the uid and full name from
        the LDAP server to use in constructing this class.

        We use
        :py:attr:`nginx_ldap_auth.settings.Settings.ldap_username_attribute` to
        determine which LDAP attribute on ``data`` holds our :py:attr:`uid` value, and
        :py:attr:`nginx_ldap_auth.settings.Settings.ldap_full_name_attribute` to
        determine which LDAP attribute holds our :py:attr:`full_name` value.

        Args:
            data: the raw LDAP data

        Returns:
            A configured :py:class:`User` object

        """
        settings = Settings()
        username_attribute = settings.ldap_username_attribute
        fullname_attribute = settings.ldap_full_name_attribute
        kwargs = {
            "uid": data[username_attribute][0],
            "full_name": data[fullname_attribute][0],
        }
        logger.info("user.parse_ldap", **kwargs)
        return cls(**kwargs)


UserManager.model = User



#!/usr/bin/env python
import sys

import click

import nginx_ldap_auth


@click.group(invoke_without_command=True)
@click.option(
    "--version/--no-version",
    "-v",
    default=False,
    help="Print the current version and exit.",
)
@click.pass_context
def cli(_, version: bool) -> None:  # noqa: FBT001
    """
    The nginx_ldap_auth command line interface.
    """
    if version:
        click.echo(nginx_ldap_auth.__version__)
        sys.exit(0)



__version__: str = "2.4.2"
__author__: str = "Caltech IMSS ADS"
__author_email__: str = "imss-ads-staff@caltech.edu"
__description__: str = (
    "A FastAPI app that authenticates users via LDAP and sets a cookie for nginx"
)



import os
import pprint

import click
import uvicorn

from ..settings import Settings
from .cli import cli

settings = Settings()


@cli.command("settings", short_help="Print our application settings.")
def print_settings():
    """
    Print our settings to stdout.  This should be the completely evaluated
    settings including those imported from any environment variable.
    """
    pp = pprint.PrettyPrinter(indent=2)
    pp.pprint(settings.model_dump())


@cli.command("start", short_help="Start the nginx_ldap_auth service.")
@click.option(
    "--host",
    "-h",
    default=lambda: os.environ.get("HOSTNAME", "0.0.0.0"),  # noqa: S104
    help="The host to listen on.",
)
@click.option(
    "--port",
    "-p",
    default=lambda: int(os.environ.get("PORT", "8888")),
    type=int,
    help="The port to listen on.",
)
@click.option(
    "--reload/--no-reload",
    "-r",
    default=lambda: os.environ.get("RELOAD", "False") == "True",
    type=bool,
    help="Reload the server on code changes.",
)
@click.option(
    "--keyfile",
    "-k",
    default=lambda: os.environ.get("SSL_KEYFILE", "/certs/server.key"),
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the SSL key file.",
)
@click.option(
    "--certfile",
    "-c",
    default=lambda: os.environ.get("SSL_CERTFILE", "/certs/server.crt"),
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the SSL certificate file.",
)
@click.option(
    "--insecure",
    default=lambda: os.environ.get("INSECURE", "False") == "True",
    type=bool,
    help="If the server should run over HTTP instead of HTTPS.",
)
@click.option(
    "--workers",
    "-w",
    default=lambda: int(os.environ.get("WORKERS", "1")),
    type=int,
    help="The number of worker processes to spawn.",
)
@click.option(
    "--env-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the environment file to load.",
)
def start(**kwargs):
    """
    Start the nginx_ldap_auth service.
    """
    uvicorn_kwargs = {
        "host": kwargs["host"],
        "port": kwargs["port"],
        "reload": kwargs["reload"],
        "workers": kwargs["workers"],
    }
    if not kwargs["insecure"]:
        # adding in SSL settings results in `uvicorn` running over HTTPS
        # the SSL settings will be ignored when insecure mode is enabled
        ssl_kwargs = {
            "ssl_keyfile": kwargs["keyfile"],
            "ssl_certfile": kwargs["certfile"],
            "ssl_version": 2,
        }
        uvicorn_kwargs |= ssl_kwargs
    if kwargs["env_file"]:
        uvicorn_kwargs["env_file"] = kwargs["env_file"]
    uvicorn.run("nginx_ldap_auth.app.main:app", **uvicorn_kwargs)



import time
from typing import Any

import bonsai
from bonsai.asyncio import AIOConnectionPool, AIOLDAPConnection
from bonsai.pool import ClosedPool, EmptyPool

from .logging import logger
from .settings import Settings


class TimeLimitedAIOLDAPConnection(AIOLDAPConnection):
    """
    A time-limited LDAP connection.  This allows us to have a connection pool
    that will close connections after a certain amount of time.

    Args:
        client: The LDAP client.

    Keyword Args:
        expires: The number of seconds after which the connection will expire.
        loop: The asyncio event loop.

    """

    def __init__(self, client: bonsai.LDAPClient, expires: int = 20, loop=None) -> None:
        super().__init__(client, loop=loop)
        self.expires = expires
        self.create_time = time.time()

    @property
    def is_expired(self) -> bool:
        return (time.time() - self.create_time) > self.expires


class TimeLimitedAIOConnectionPool(AIOConnectionPool):
    """
    A pool of time-limited LDAP connections.  This allows us to have relatively
    fresh connections to our LDAP server while not having to create a new
    connection for every request.

    Args:
        settings: The application settings.
        client: The LDAP client.

    Keyword Args:
        minconn: The minimum number of connections to keep in the pool.
        maxconn: The maximum number of connections to keep in the pool.
        loop: The asyncio event loop.

    """

    def __init__(
        self,
        settings: Settings,
        client: bonsai.LDAPClient,
        minconn: int = 1,
        maxconn: int = 10,
        loop=None,
        **kwargs: Any,
    ) -> None:
        super().__init__(client, minconn, maxconn, loop=loop, **kwargs)
        self.settings = settings

    async def get(self) -> AIOLDAPConnection:  # type: ignore[override]
        """
        Get a connection from the pool.  If a connection has expired, close it
        and create a new connection, then return the new connection.

        Raises:
            ClosedPool: The pool has not been initialized.
            EmptyPool: There are no connections in the pool.

        Returns:
            A connection from the pool.

        """
        async with self._lock:
            if self._closed:
                msg = "The pool is closed."
                raise ClosedPool(msg)
            await self._lock.wait_for(lambda: not self.empty or self._closed)
            try:
                conn = self._idles.pop()
            except KeyError:
                if len(self._used) < self._maxconn:
                    conn = await self._client.connect(
                        is_async=True, loop=self._loop, **self._kwargs
                    )
                else:
                    msg = "Pool is empty."
                    raise EmptyPool(msg) from None
            if conn.is_expired:
                logger.info(
                    "ldap.pool.connection.recycle",
                    lifetime_seconds=self.settings.ldap_pool_connection_lifetime_seconds,
                )
                # Does this need to be awaited?
                conn.close()
                conn = await self._client.connect(
                    is_async=True, loop=self._loop, **self._kwargs
                )
            self._used.add(conn)
            self._lock.notify()
            return conn



# Configuration file for the Sphinx documentation builder.  # noqa: INP001
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))  # noqa: PTH100

from typing import Any

import sphinx_rtd_theme  # pylint: disable=unused-import  # noqa:F401

# These are required for sphinx-apidoc to work
os.environ["LDAP_URI"] = "ldap://ldap.example.com"
os.environ["LDAP_BINDDN"] = "cn=admin,dc=example,dc=com"
os.environ["LDAP_PASSWORD"] = "password"  # noqa: S105
os.environ["LDAP_BASEDN"] = "dc=example,dc=com"
os.environ["SECRET_KEY"] = "my-key"  # noqa: S105

# -- Project information -----------------------------------------------------

# the master toctree document
master_doc: str = "index"

project: str = "nginx-ldap-auth-service"
copyright: str = "2023, Caltech IMSS ADS"  # noqa: A001
author: str = "Caltech IMSS ADS"

# The full version, including alpha/beta/rc tags
release: str = "2.4.2"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions: list[str] = [
    "sphinx.ext.autodoc",
    "sphinxcontrib.images",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_theme",
    "sphinx_json_globaltoc",
]

source_suffix: str = ".rst"

# Add any paths that contain templates here, relative to this directory.
templates_path: list[str] = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: list[str] = ["_build"]

add_function_parentheses: bool = False
add_module_names: bool = True

autodoc_member_order: str = "bysource"
autodoc_type_aliases: dict[str, str] = {}

# the locations and names of other projects that should be linked to this one
intersphinx_mapping: dict[str, tuple[str, str | None]] = {
    "python": ("https://docs.python.org/3", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme: str = "sphinx_rtd_theme"

html_show_sourcelink: bool = False
html_show_sphinx: bool = False
html_show_copyright: bool = True
html_theme_options: dict[str, Any] = {"collapse_navigation": False}



import logging
import logging.config
import sys
from typing import TYPE_CHECKING, Any, cast

import structlog
from fastapi import Request

from nginx_ldap_auth.settings import Settings

if TYPE_CHECKING:
    from starlette.datastructures import Address

settings = Settings()
logger = structlog.get_logger("nginx_ldap_auth")


def get_logger(request: Request | None = None) -> structlog.BoundLogger:
    """
    Return a structlog logger with request context information.

    Creates a logger that includes relevant request context when available, making
    it easier to trace logs related to specific requests.

    Note:
        When a request is provided, the logger is bound with the following values:

        * realm: The authentication realm (from x-auth-realm header or settings)
        * host: The server hostname (from host header)
        * remote_ip: The client IP address (from x-forwarded-for or request.client)


    Args:
        request: The FastAPI/Starlette request object. If provided, adds request
                context information to the logger.

    Returns:
        structlog.BoundLogger: A configured structlog logger with optional
        request context

    """
    if request is None:
        return logger
    remote_ip = request.headers.get("x-forwarded-for")
    if remote_ip:
        remote_ip = remote_ip.split(",")[0]
    else:
        remote_ip = cast("Address", request.client).host
    request = cast("Request", request)
    return logger.bind(
        realm=request.headers.get("x-auth-realm", settings.auth_realm),
        host=request.headers.get("host", "unknown"),
        remote_ip=remote_ip,
    )


class ContextLoggingProcessor:
    """
    Structlog processor that adds additional context to log events.

    This processor allows you to define default key-value pairs that will be
    added to all log events processed by structlog. Values provided explicitly
    during logging will take precedence over these defaults.

    Attributes:
        log_kwargs: Dictionary of key-value pairs to add to log events

    Example:
        To make all log events include ``app_name`` and ``environment`` unless
        explicitly overridden:

        .. code-block:: python

            import structlog
            from nginx_ldap_auth.logging import ContextLoggingProcessor

            # Create a processor that adds app_name and environment to all log events
            processor = ContextLoggingProcessor(
                app_name="my-app",
                environment="production"
            )

    """

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the processor with context values.

        Args:
            **kwargs: Key-value pairs to add to all log events

        """
        self.log_kwargs = kwargs

    def __call__(self, _, __, event_dict):
        """
        Add context values to the log event dictionary.

        This processor adds all key-value pairs from the initialization to the
        event dictionary, but doesn't overwrite any existing values.

        Args:
            _: Logger (unused)
            __: Method name (unused)
            event_dict: The log event dictionary to modify

        Returns:
            dict: The modified log event dictionary

        """
        for k, v in self.log_kwargs.items():
            event_dict.setdefault(k, v)
        return event_dict


class CensorPasswordProcessor:
    """
    Structlog processor that censors password fields in log events.

    This processor automatically detects and censors common password field names
    in log events, preventing sensitive information from being logged.

    Attributes:
        log_kwargs: Additional configuration options (unused currently)

    Example:
        >>> # Add to structlog processors
        >>> processors = [
        >>>     # other processors
        >>>     CensorPasswordProcessor(),
        >>>     # more processors
        >>> ]

    """

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the processor.

        Args:
            **kwargs: Configuration options (reserved for future use)

        """
        self.log_kwargs = kwargs

    def __call__(self, _, __, event_dict):
        """
        Censor password fields in the log event dictionary.

        Automatically identifies and censors values for keys named "password",
        "password1", or "password2" to prevent sensitive data from being logged.

        Args:
            _: Logger (unused)
            __: Method name (unused)
            event_dict: The log event dictionary to modify

        Returns:
            dict: The modified log event dictionary with passwords censored

        """
        for password_key_name in ("password", "password1", "password2"):
            if password_key_name in event_dict:
                event_dict[password_key_name] = "*CENSORED*"
        return event_dict


structlog.configure(
    processors=[
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        ContextLoggingProcessor(),
        CensorPasswordProcessor(),
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ],
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=False,
)

pre_chain = [
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.processors.TimeStamper(fmt="iso"),
]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "handlers": {
        "default": {
            "level": settings.loglevel,
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
        "access": {"class": "logging.StreamHandler", "formatter": "access"},
    },
    "loggers": {
        "uvicorn.error": {
            "level": settings.loglevel,
            "handlers": ["default"],
            "propagate": False,
        },
        "uvicorn.access": {
            "level": settings.loglevel,
            "handlers": ["access"],
            "propagate": False,
        },
    },
    "root": {
        # Set up the root logger.  This will make all otherwise unconfigured
        # loggers log through structlog processor.
        "handlers": ["default"],
        "level": settings.loglevel,
    },
    "formatters": {
        "default": {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.processors.JSONRenderer(),
            "foreign_pre_chain": pre_chain,
            "format": "%(message)s",
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "format": "%(asctime)s %(message)s",
        },
    },
}


if settings.log_type == "text":
    LOGGING["formatters"]["default"]["processor"] = structlog.dev.ConsoleRenderer()  # type: ignore[index]


logging.config.dictConfig(LOGGING)


def handle_exception(exc_type, exc_value, exc_traceback):
    """
    Log any uncaught exception instead of letting it be printed by Python
    (but leave KeyboardInterrupt untouched to allow users to Ctrl+C to stop)
    See https://stackoverflow.com/a/16993115/3641865
    """
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = handle_exception
